#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <ctime>
#include <cstring>
#include <cctype>

int usage();

int main(int argc, char *argv[])
{
	//Parsing.
	if (argc<6  || strcmp(argv[2], "{") !=0)
		return usage();
	char genomesize[50], basename[200], fullname[200], pair_end=1, dump_num=0, dump_name[100];
	int max_reads_displayed=1, mismatches_allowed, alignments_allowed=5, max_read_length=1000, min_pHRED=3, stop, n=3;
	int head_times=0, max_head_times=100, max_insert_size=1000;
	unsigned long long int genome;
	bool set_mismatches=false, set_genome=false, dump_unmatched=false;
	while (strcmp(argv[n], "}") !=0 && n<argc)
		n++;
	if (n>=argc-1)
		return usage();
	else	stop=n;
	n+=2;
	for (n; n<argc; n++)
	{
		if 	(strcmp(argv[n], "-g")==0)
		{
			set_genome=true;
			n++;
			strcpy(genomesize,argv[n]);
			genome=strtoull(genomesize, NULL, 0);
		}
		else if (strcmp(argv[n], "-q")==0)
		{
			n++;
			min_pHRED=atoi(argv[n]);
		}
		else if	(strcmp(argv[n], "-r")==0)
		{
			n++;
			max_read_length=atoi(argv[n]);
		}
		else if (strcmp(argv[n], "-p")==0)
		{
			n++;
			max_reads_displayed=atoi(argv[n]);
		}
		else if (strcmp(argv[n],"-m")==0)
		{
			n++;
			mismatches_allowed=atoi(argv[n]);
			set_mismatches=true;
		}
		else if (strcmp(argv[n], "-a")==0)
		{
			n++;
			alignments_allowed=atoi(argv[n]);
		}
		else if (strcmp(argv[n], "-s")==0)
		{
			n++;
			max_head_times=atoi(argv[n]);
			if (max_head_times==-1)
				max_head_times=2000000000;
		}
		else if (strcmp(argv[n], "-P")==0)
			pair_end=2;
		else if (strcmp(argv[n], "-i")==0)
		{
			n++;
			max_insert_size=atoi(argv[n]);
		}
		else if (strcmp(argv[n], "-f")==0)
		{
			dump_unmatched=true;
			n++;
			dump_num=n;
		}
			
 		else		
 			return usage();
 	}
	const int ALLOWED=alignments_allowed, MRL=max_read_length+1;
	const char PAIRS=pair_end;
	
	FILE *index_header, *bfa, *fastq, *fastq2, *index, *index_val, *alignment, *ACGT_alignment[4][PAIRS], *fq_out[PAIRS];
	
	//Open and test input and output files.
	index=fopen(argv[1], "rb");
	if (index == (FILE*)0)
	{
		fprintf(stderr,"*************************\n");
		fprintf(stderr,"Error opening .sqn file*\n");
		fprintf(stderr,"*************************\n");
		return usage();
	}
	strcpy(basename, argv[1]);
	strcpy(fullname,basename);
	strcat(fullname,".hdr");
	index_header = fopen(fullname, "r");
	if (index_header == (FILE*)0)
	{
		fprintf(stderr,"*****************************\n");
		fprintf(stderr,"Error opening .sqn.hdr file*\n");
		fprintf(stderr,"*****************************\n");
		return usage();
	}
	strcpy(fullname,basename);
	strcat(fullname,".bfa");
	bfa = fopen(fullname, "rb");
	if (bfa == (FILE*)0)
	{
		fprintf(stderr,"*****************************\n");
		fprintf(stderr,"Error opening .sqn.bfa file*\n");
		fprintf(stderr,"*****************************\n");
		return usage();
	}
	strcpy(fullname,basename);
	strcat(fullname,".val");
	index_val = fopen(fullname, "rb");
	if (index_val == (FILE*)0)
	{
		fprintf(stderr,"*****************************\n");
		fprintf(stderr,"Error opening .sqn.val file*\n");
		fprintf(stderr,"*****************************\n");
		return usage();
	}
	for (n=3; n<stop; n++)
	{
		fastq=fopen64(argv[n], "r");
		if (fastq == (FILE*)0)
		{
			fprintf(stderr,"**************************\n");
			fprintf(stderr,"Error opening .fastq file %s\n", argv[n]);
			fprintf(stderr,"**************************\n");
			return usage();
		}
		fclose(fastq);
	}
	alignment=fopen64(argv[stop+1], "w");
	if (alignment == (FILE*)0)
	{
		fprintf(stderr,"**************************\n");
		fprintf(stderr,"Error opening .sam file*\n");
		fprintf(stderr,"**************************\n");
		return usage();
	}
	strcpy(basename, argv[stop+1]);
	if (dump_unmatched==true)
	{
		strcpy(dump_name, argv[dump_num]);
		if (PAIRS==1)
		{
			strcat (dump_name, ".fastq");
			fq_out[0]=fopen64(dump_name, "w");
		}
		else
		{
			strcpy(fullname, dump_name);
			strcat(fullname, "_1.fastq");
			fq_out[0]=fopen64(fullname, "w");
			strcpy(fullname, dump_name);
			strcat(fullname, "_2.fastq");
			fq_out[1]=fopen64(fullname, "w");
		}
	}
	//Store chromosome number and height and size of hash table, and write to .SAM file.
	int header,filenamepos=0, zeros[4], strchar=0;
	unsigned long int total_length=0;
	char string[1000], filename[200];
	const unsigned long int HEIGHT=strtoul(fgets(string,100,index_header),NULL,0);
	const unsigned long int SIZE=strtoul(fgets(string,100,index_header), NULL,0);
	const unsigned long int NUM_OF_CHR=strtoul(fgets(string,100,index_header), NULL,0);
	int chr_ref_len[NUM_OF_CHR];
	char chr_name[NUM_OF_CHR][1000];
	for (int count=0; count<NUM_OF_CHR; count++)
	{
		fgets(filename,200, index_header);
		fgets(string,1000, index_header);
		fprintf(alignment, "@SQ\tSN:");
		strchar=0;
		while (isspace(string[strchar])==false)
		  {
		    fprintf(alignment,"%c",string[strchar]);
		    strchar++;
		  }
		string[strchar]='\0';
		strcpy(chr_name[count], string);
		chr_ref_len[count]=atoi(fgets(string,1000,index_header));
		fprintf(alignment,"\tLN:%u\n", chr_ref_len[count]);
		total_length=total_length+chr_ref_len[count];
	}
	const unsigned long int REF_LEN=total_length;
	for (int count=0; count<4; count++)
		zeros[count]=atoi(fgets(string,100, index_header));
	const unsigned long int NUM_OF_VALS=strtoul(fgets(string,100,index_header),NULL,0);
	
	//Variable declarations 
	int is_amp, pHRED, pHRED_table[max_read_length+1][100], reads_processed=0, read_length, rl2, end, start, p=0, q=0, b, c;
	int mismatches=0, to_print, temp_mismatches_allowed, alignment_num, prints=0, rand_read, max_mismatches_allowed[max_read_length+1];
	char exclude[200], QNAME[100], qualities[MRL],seq[MRL], *mismatch_holder[1000];
	char QNAME2[100], qualities2[MRL], seq2[MRL];
	unsigned int x=0, base_num, sequence[MRL], read_loc, stored[2][alignments_allowed],loc_in_genome=0, stop_number=0, spot=0;
	unsigned int power[HEIGHT][4],mismatch_spot=0;
	bool found=false, already_found=false;
	srand(time(0));
	char ACGT_names[4][PAIRS][200], matched_pairs[4*ALLOWED][2], match_num=0;
	unsigned int ACGT_locs[4][PAIRS][ALLOWED], ACGT_strand[4][PAIRS][ALLOWED], ACGT_mismatches[4][PAIRS], ACGT_hits[4][PAIRS];
	unsigned int total_hits[2], all_locs[3][PAIRS][4*ALLOWED];
	
	c=1;
	for (b=0; b<HEIGHT; b++)
	{
		power[b][0]=0;
		power[b][1]=c;
		power[b][2]=2*c;
		power[b][3]=3*c;
		c=4*c;
	}
	b=c=0;
	
	//Read the .bfa file into memory.
	const int M_NUM_OF_VALS=(NUM_OF_VALS/1000000+1), R_NUM_OF_VALS=NUM_OF_VALS%1000000;
	const int M_SIZE=(SIZE/1000000+1), R_SIZE=SIZE%1000000;
	unsigned char *fasta_locations[(REF_LEN/4000000)+1];
	char base;
	int i=0, j=0;
	for (int a=0; a<(REF_LEN/4000000+1); a++)
	{
		fasta_locations[a]= new unsigned char[1000001];
		fread(fasta_locations[a],1,1000000, bfa);
	}
	//build quality table
	unsigned long long int tot=1, tot2=1, com1=1, com2=0, com3=1, com4=2*REF_LEN;
	int t1=0, t2=0, c1=0, c2=0, c3=0, c4=0, mm1=1;
	while (com4>1000llu)
	{
		com4=com4/10;
		c4++;
	}
	if (set_mismatches==false)
	{
		for (int v=1; v<=max_read_length; v++)
		{
			com1=1;
			com2=0;
			com3=1;
			mm1=1;
			c1=c2=c3=0;
			tot=tot*4;
			if (tot>1000000000000000llu)
			{
				tot=tot/100000;
				t1+=5;
			}
			tot2=tot;
			t2=t1;
			while (tot2>1000llu)
			{
				tot2=tot2/10;
				t2++;
			}
			if (com4>=tot2)
				pHRED = t2-c4-1;
			else	pHRED = t2-c4;
			pHRED_table[v][0]=pHRED;
			while (pHRED>min_pHRED  && mm1<100)
			{
				com1=com1*3*(v+1-mm1)/mm1;
				com2=com2+com1;
				if (com2>1000000000000000llu)
				{
					com1=com1/100000;
					c1+=5;
					com2=com2/100000;
					c2+=5;
				}
				c3=c2;
				com3=com2;
				while (com3>1000llu)
				{
					com3=com3/10;
					c3++;
				}
				com3=com3*REF_LEN;
				while (com3>1000llu)
				{
					com3=com3/10;
					c3++;
				}
				if (com3>=tot2)
					pHRED=t2-c3-1;
				else	pHRED=t2-c3;
				pHRED_table[v][mm1]=pHRED;
				mm1++;
 			}
 			max_mismatches_allowed[v]=mm1-1;
		}
	}
	else 
	{
		for (int v=1; v<=max_read_length; v++)
		{
			com1=1;
			com2=0;
			com3=1;
			mm1=1;
			c1=c2=c3=0;
			tot=tot*4;
			if (tot>1000000000000000llu)
			{
				tot=tot/100000;
				t1+=5;
			}
			tot2=tot;
			t2=t1;
			while (tot2>1000llu)
			{
				tot2=tot2/10;
				t2++;
			}
			if (com4>=tot2)
				pHRED = t2-c4-1;
			else	pHRED = t2-c4;
			pHRED_table[v][0]=pHRED;
			while (pHRED>min_pHRED  && mm1<=mismatches_allowed)
			{
				com1=com1*3*(v+1-mm1)/mm1;
				com2=com2+com1;
				if (com2>1000000000000000llu)
				{
					com1=com1/100000;
					c1+=5;
					com2=com2/100000;
					c2+=5;
				}
				c3=c2;
				com3=com2;
				while (com3>1000llu)
				{
					com3=com3/10;
					c3++;
				}
				com3=com3*REF_LEN;
				while (com3>1000llu)
				{
					com3=com3/10;
					c3++;
				}
				if (com3>=tot2)
					pHRED=t2-c3-1;
				else	pHRED=t2-c3;
				pHRED_table[v][mm1]=pHRED;
				mm1++;
			}
			max_mismatches_allowed[v]=mm1-1;
			
		}
	}
	//Read the index file into memory.
	unsigned int *index_vals[M_SIZE];
	unsigned int *locations[M_NUM_OF_VALS];	//Split the arrays to hold the locations and array
	unsigned int size=0;						//coordinates of the first location for each leaf into
	unsigned int head_num=0, head_loc=0;				//fragments that are small enough to be read into memory
	char *is_new_line=new char;					//Get the white space character between each location.	//ie:
	unsigned char val_in_genome=0;					//make an array of pointers to arrays of locations or 
	for (int a=0; a<M_NUM_OF_VALS-1; a++)				//coordinates.
		locations[a]= new unsigned int[1000000];		//The arrays of pointers are the respective sizes divided
	locations[M_NUM_OF_VALS-1]=new unsigned int[R_NUM_OF_VALS];	//by 1 million so each array should hold 1 million locations
	for (int a=0; a<M_SIZE-1; a++)					//or coordinates.	
		index_vals[a]=new unsigned int [1000000];

	index_vals[M_SIZE-1]=new unsigned int[R_SIZE+1];
	for (int a=0; a<M_NUM_OF_VALS-1; a++)
		fread(locations[a],4,1000000, index);
	fread(locations[M_NUM_OF_VALS-1],4,R_NUM_OF_VALS,index);
	
for (int first_base=0; first_base<4; first_base++)
{
	strcpy(fullname,basename);
	if 	(first_base==0)
		strcat(fullname, ".A");
	else if	(first_base==1)
		strcat(fullname, ".C");
	else if	(first_base==2)
		strcat(fullname, ".G");
	else	strcat(fullname, ".T");
	ACGT_alignment[first_base][0]=fopen64(fullname, "wb");

	if (PAIRS==2)
	{
		strcat(fullname, ".2");
		ACGT_alignment[first_base][1]=fopen64(fullname, "wb");
	}
		
	
	for (int a=0; a<M_SIZE-1; a++)
		fread(index_vals[a],4,1000000, index_val);
	fread(index_vals[M_SIZE-1],4,R_SIZE,index_val);
	//Process all short reads
	for(n=3; n<stop; n++)
	{
	pair_end=(n+1)%2;
 	fastq=fopen64(argv[n], "r");
 	while (is_amp=fgetc(fastq) != EOF)
 	{
	  //	  fprintf(stdout, "temp mm=%d\n", temp_mismatches_allowed);
		if ((reads_processed/4000000)==mismatch_spot)
		{
			mismatch_holder[mismatch_spot]=new char [4000000];
			mismatch_spot++;
		}
		x=0;
		QNAME[x]=fgetc(fastq);
		while (QNAME[x] != ' ' && QNAME[x] != '\t' && QNAME[x] != '\n')
		{					//First line of a short read:
			x++;				//Store the sequence name in 
			QNAME[x]=fgetc(fastq);		//QNAME and discard the rest 
		}					//of the information from the
		if (QNAME[x]!='\n')			//first line if any additional
			fgets(exclude, MRL, fastq);	//information exists.
		QNAME[x]='\0';
		x=0;
		seq[x]=fgetc(fastq);			//Line 2: sequence of short read
		while (seq[x] != '\n')			//Read in sequence and convert 
		{					//bases to numerical values
			switch (seq[x])
			{
				case 'A':
					sequence[x]=0;
					break;
				case 'C':
					sequence[x]=1;
					break;
				case 'G':
					sequence[x]=2;
					break;
				case 'T':
					sequence[x]=3;
					break;
				default :
					sequence[x]=rand()%4;
			}
			x++;
			seq[x]=fgetc(fastq);
		}
		read_length=x;
		fgets(exclude,MRL, fastq);		//Ignore Line 3:
		fgets(qualities,MRL, fastq);		//Line 4: Read qualities.
		alignment_num=0;
		found=false;
		if (set_mismatches==false)
		{
			if (first_base==0)
				temp_mismatches_allowed=max_mismatches_allowed[read_length];
			else
				temp_mismatches_allowed=*(mismatch_holder[reads_processed/4000000]+reads_processed%4000000);
		}
		else
		{
			if (first_base==0)
				temp_mismatches_allowed=mismatches_allowed;
			else 
				temp_mismatches_allowed=*(mismatch_holder[reads_processed/4000000]+reads_processed%4000000);
		}
		for (end=HEIGHT; (end<=read_length && end<(temp_mismatches_allowed+2)*HEIGHT); end++)
		{
			
			if (sequence[end-HEIGHT]==first_base)
			{
			read_loc=0;
			start=end-HEIGHT;
			for (x=1; x<HEIGHT; x++)
				read_loc=(read_loc+power[x-1][sequence[end-x]]);	//Get leaf number
			stop_number=*(index_vals[read_loc/1000000]+read_loc%1000000);//Get the location element from coordinate array.
			if(read_loc !=0)
				head_loc=*(index_vals[(read_loc-1)/1000000]+(read_loc-1)%1000000);
			else	head_loc=zeros[first_base];
			while (head_loc < stop_number && head_times <max_head_times)	//Read in from the locations array until we have read to
 			{								//the stop number element.
				loc_in_genome=*(locations[head_loc/1000000]+(head_loc%1000000));
				spot=loc_in_genome+HEIGHT-1;
				i=(loc_in_genome+HEIGHT-1)/4000000;
				j=(loc_in_genome+HEIGHT-1)%4000000;
  				mismatches=0;
 				q=0;
				while ((end+q) < read_length && mismatches<=temp_mismatches_allowed)
 				{
 					b=(spot+q)%4;
					val_in_genome=*(fasta_locations[i]+(j/4));
					j+=(4-b);
					if (b==0)
					{
 						if (sequence[end+q] != (val_in_genome/64))
						  mismatches++;
						b++;
						q++;
					}
					if (b==1 && (end+q) < read_length && mismatches<=temp_mismatches_allowed)
 					{
						if (sequence[end+q] != (val_in_genome%64)/16)
						  mismatches++;
						b++;
						q++;
					}
					if (b==2 && (end+q) < read_length && mismatches<=temp_mismatches_allowed)
					{
 						if (sequence[end+q] != (val_in_genome%16)/4)
							mismatches++;
						b++;
						q++;
					}
					if (b==3 && (end+q) < read_length && mismatches<=temp_mismatches_allowed)
					{
 						if (sequence[end+q] != val_in_genome%4)
							mismatches++;
						b++;
						q++;
					}
					if (j>=4000000)
					{
						j-=4000000;
						i++;
					}
						
 				}
 				if (loc_in_genome>start)
 				{
					q=0;
					spot=loc_in_genome-start-1;
					i=(loc_in_genome-start-1)/4000000;
					j=(loc_in_genome-start-1)%4000000;
  					while (q<start && mismatches<=temp_mismatches_allowed)
  					{
						val_in_genome=*(fasta_locations[i]+(j/4));
 						b=(spot+q)%4;
						j+=4-b;
						if (b==0 && mismatches<=temp_mismatches_allowed)
						{
							if (sequence[q] != val_in_genome/64)
								mismatches++;
							b++;
							q++;
						}
						if (b==1 && q<start && mismatches<=temp_mismatches_allowed)
						{
							if (sequence[q] != (val_in_genome%64)/16)
								mismatches++;
							b++;
							q++;
						}
						if (b==2 && q<start && mismatches<=temp_mismatches_allowed)
						{
							if (sequence[q] != (val_in_genome%16)/4)
								mismatches++;
							b++;
							q++;
						}
						if (b==3 && q<start && mismatches<=temp_mismatches_allowed)
						{
							if (sequence[q] != val_in_genome%4)
								mismatches++;
							b++;
							q++;
						}
						if (j>=4000000)
						{
							j-=4000000;
							i++;
						}
 					}
				}
				else
					mismatches=temp_mismatches_allowed+1;
				//Do not search for alignments with more mismatches than the alignment with the fewest mismatches
				if (mismatches<temp_mismatches_allowed)
				{
					temp_mismatches_allowed=mismatches;
					alignment_num=0;//Each time an alignment with fewer mismatches is found, begin storing at the top of list.
				}	// Overwriting reads with more mismatches is ok sicne only the best alignment will be displayed.
			
				//Store the matches on the lead strand.
				if (mismatches<=temp_mismatches_allowed)
				{
					already_found=false;
					for (int h=0; h<alignment_num && already_found==false; h++)
						if (loc_in_genome-start==stored[h][0] )
							already_found=true;
					if (already_found==false)
					{
						stored[0][alignment_num]=loc_in_genome-start;//alignment_location
						stored[1][alignment_num]=0;	//flag
						alignment_num++;
						found=true;
					}
				}
				if (alignment_num==alignments_allowed)
				{
					if (temp_mismatches_allowed !=0)	//If "alignments_allowed" with a certain number of mismatches have
					{					//been found, only search for alignments with fewer mismatches.
						temp_mismatches_allowed--;
						alignment_num=0;
					}
					else 
					{
						head_loc=stop_number; 	//If "alignemnts_allowed" hits with zero mismathces have been found, 
						end=read_length;	//terminate for loop.
					}
				}
				head_loc++;
				head_times++;
			}
			
 			}
			
 		}
 
 		//Repeat for the reverse strand .
		if (temp_mismatches_allowed!=0 || alignment_num!=alignments_allowed) 	//If "alignements_allowed" hits with zero mismatches
 		{									//have been found, don't search on reverse strand.
 			for (end=HEIGHT; (end<=read_length && end<(temp_mismatches_allowed+2)*HEIGHT); end++)
 			{
				if ((3-sequence[end-1])==first_base)
				{
 				start=end-HEIGHT;
 				read_loc=0;
 				for (x=0; x<(HEIGHT-1); x++)
 					read_loc=(read_loc+power[x][3-sequence[start+x]]);
				stop_number=*(index_vals[read_loc/1000000]+read_loc%1000000);//Get the location element from coordinate array.
				if(read_loc !=0)
					head_loc=*(index_vals[(read_loc-1)/1000000]+(read_loc-1)%1000000);
				else	head_loc=zeros[first_base];
				head_times=0;
				while (head_loc < stop_number && head_times<max_head_times)
 				{
					loc_in_genome=*(locations[head_loc/1000000]+(head_loc%1000000));

 					mismatches=0;
					spot=loc_in_genome-2;
					i=(loc_in_genome-2)/4000000;
					j=(loc_in_genome-2)%4000000;
 					if (loc_in_genome > read_length-end)
 					{
 						q=0;
 						while ((end+q) < read_length && mismatches<=temp_mismatches_allowed)
						{
							val_in_genome=*(fasta_locations[i]+(j/4));
							b=(spot-q)%4;
							j-=(1+b);
							if (b==3)
							{
								if ((3-sequence[end+q]) != val_in_genome%4)
									mismatches++;
								b--;
								q++;
							}
							if (b==2 && (end+q) < read_length && mismatches<=temp_mismatches_allowed)
							{
								if ((3-sequence[end+q]) != (val_in_genome%16)/4)
									mismatches++;
								b--;
								q++;
							}
							if (b==1 && (end+q) < read_length && mismatches<=temp_mismatches_allowed)
							{
								if ((3-sequence[end+q] )!= (val_in_genome%64)/16)
									mismatches++;
								b--;
								q++;
							}
							if (b==0 && (end+q) < read_length && mismatches<=temp_mismatches_allowed)
							{
								if ((3-sequence[end+q]) != val_in_genome/64)
									mismatches++;
								b--;
								q++;
							}
							if (j<0)
							{
								i--;
								j+=4000000;
							}
						}
					}
					else  
						mismatches=temp_mismatches_allowed+1;
 	 				q=1;
					spot=loc_in_genome+HEIGHT-2;
					i=(loc_in_genome+HEIGHT-1)/4000000;
					j=(loc_in_genome+HEIGHT-1)%4000000;
	 				while (q<=start && mismatches<=temp_mismatches_allowed)
	 				{
	 					val_in_genome=*(fasta_locations[i]+(j/4));
						b=(spot+q)%4;
						j+=4-b;
						if (b==0)
						{
							if ((3-sequence[start-q]) != val_in_genome/64)
								mismatches++;
							b++;
							q++;
						}
						if (b==1 && q<=start && mismatches<=temp_mismatches_allowed)
						{
							if ((3-sequence[start-q] )!= (val_in_genome%64)/16)
								mismatches++;
							b++;
							q++;
						}
						if (b==2 && q<=start && mismatches<=temp_mismatches_allowed)
						{
							if ((3-sequence[start-q]) != (val_in_genome%16)/4)
								mismatches++;
							b++;
							q++;
						}
						if (b==3 && q<=start && mismatches<=temp_mismatches_allowed)
						{
							if ((3-sequence[start-q]) != val_in_genome%4)
								mismatches++;
							b++;
							q++;
						}
						if (j>=4000000)
						{
							j-=4000000;
							i++;
						}
	 				}
	 				if (mismatches<temp_mismatches_allowed)
	 				{
	 					temp_mismatches_allowed=mismatches;
	 					alignment_num=0;//Each time an alignment with fewer mismatches is found, begin 
								//storing at the top of list.
	 				}	// Overwriting reads with more mismatches is ok sicne only the best alignment will be displayed.
	 				//Store the matches on the reverse strand.
	 				if (mismatches<=temp_mismatches_allowed)
	 				{
	 					already_found=false;
	 					for (int h=0; h<alignment_num & already_found==false; h++)
	 						if (loc_in_genome-read_length+end==stored[h][0] )
	 							already_found=true;
						if (already_found==false)
						{
	 						stored[0][alignment_num]=loc_in_genome-read_length+end;	//alignemnt locations
	 						stored[1][alignment_num]=16;				//flag.
	 						alignment_num++;
							found=true;
						}
	 				}
	 				if (alignment_num==alignments_allowed)
	 				{
	 					if (temp_mismatches_allowed !=0)
	 					{
	 						temp_mismatches_allowed--;
	 						alignment_num=0;
	 					}
	 					else 
						{
							head_loc=stop_number;
							end=read_length;
						}
	 				}
					head_loc++;
					head_times++;
	 			}
				}
	 		}
 		}
		*(mismatch_holder[reads_processed/4000000]+(reads_processed%4000000))=temp_mismatches_allowed;
		//store temporary alignments;
		if 	(found==true && PAIRS==2)
		{
			if (alignment_num==0)
			{
				alignment_num=alignments_allowed;
				temp_mismatches_allowed++;
			}
			for(x=0; QNAME[x] != '\0'; x++)
				fprintf(ACGT_alignment[first_base][pair_end], "%c",QNAME[x]);
			fprintf(ACGT_alignment[first_base][pair_end], "\n");
			fwrite(&alignment_num, 1,4,ACGT_alignment[first_base][pair_end]);
			fwrite(&temp_mismatches_allowed,1,4,ACGT_alignment[first_base][pair_end]);
			fwrite(&stored[0],alignment_num,4,ACGT_alignment[first_base][pair_end]);
			fwrite(&stored[1],alignment_num,4,ACGT_alignment[first_base][pair_end]);
		}
		
		else if (found==true && PAIRS==1)
		{
			if (alignment_num==0)
			{
				alignment_num=alignments_allowed;
				temp_mismatches_allowed++;
			}
			for(x=0; QNAME[x] != '\0'; x++)
				fprintf(ACGT_alignment[first_base][0], "%c",QNAME[x]);
			fprintf(ACGT_alignment[first_base][0], "\n");
			fwrite(&alignment_num, 1,4,ACGT_alignment[first_base][0]);
			fwrite(&temp_mismatches_allowed,1,4,ACGT_alignment[first_base][0]);
			fwrite(&stored[0],alignment_num,4,ACGT_alignment[first_base][0]);
			fwrite(&stored[1],alignment_num,4,ACGT_alignment[first_base][0]);
		}	
		
		reads_processed++;
		if (reads_processed%1000000==0)
		{	
			if 	(first_base==0)
				fprintf(stdout, "Alignment A: %d reads processed.\n", reads_processed);
			else if	(first_base==1)
				fprintf(stdout, "Alignment C: %d reads processed.\n", reads_processed);
			else if	(first_base==2)
				fprintf(stdout, "Alignment G: %d reads processed.\n", reads_processed);
			else
				fprintf(stdout, "Alignment T: %d reads processed.\n", reads_processed);
		}
	}
	fclose(fastq);
	}
	if 	(first_base==0)
		fprintf(stdout, "Alignment A: %d reads processed.\n", reads_processed);
	else if	(first_base==1)
		fprintf(stdout, "Alignment C: %d reads processed.\n", reads_processed);
	else if	(first_base==2)
		fprintf(stdout, "Alignment G: %d reads processed.\n", reads_processed);
	else
		fprintf(stdout, "Alignment T: %d reads processed.\n", reads_processed);
	reads_processed=0;
	fclose(ACGT_alignment[first_base][0]);
	if (PAIRS==2)
		fclose(ACGT_alignment[first_base][1]);
 }
 	//Write to permanent file
	for (p=0; p<4; p++)
	{
		strcpy(fullname,basename);
		if 	(p==0)
			strcat(fullname, ".A");
		else if	(p==1)
			strcat(fullname, ".C");
		else if	(p==2)
			strcat(fullname, ".G");
		else	strcat(fullname, ".T");
		ACGT_alignment[p][0]=fopen(fullname, "rb");
		fgets(ACGT_names[p][0],200,ACGT_alignment[p][0]);
		fread(&ACGT_hits[p][0],4,1,ACGT_alignment[p][0]);
		fread(&ACGT_mismatches[p][0],4,1,ACGT_alignment[p][0]);
		fread(&ACGT_locs[p][0],4,ACGT_hits[p][0],ACGT_alignment[p][0]);
		fread(&ACGT_strand[p][0],4,ACGT_hits[p][0],ACGT_alignment[p][0]);
		 
		if (PAIRS==2)
		{
			strcat(fullname,".2");
			ACGT_alignment[p][1]=fopen(fullname, "rb");
			fgets(ACGT_names[p][1],200,ACGT_alignment[p][1]);
			fread(&ACGT_hits[p][1],4,1,ACGT_alignment[p][1]);
			fread(&ACGT_mismatches[p][1],4,1,ACGT_alignment[p][1]);
			fread(&ACGT_locs[p][1],4,ACGT_hits[p][1],ACGT_alignment[p][1]);
			fread(&ACGT_strand[p][1],4,ACGT_hits[p][1],ACGT_alignment[p][1]);
		}
	}
	if (PAIRS==1)
	{
	  reads_processed=0;
		for (n=3; n<stop; n++)
		{
		fastq=fopen64(argv[n], "r");
		while (is_amp=fgetc(fastq) != EOF)
		{
			x=0;
			QNAME[x]=fgetc(fastq);
			while (QNAME[x] != ' ' && QNAME[x] != '\t' && QNAME[x] != '\n')
			{					//First line of a short read:
				x++;				//Store the sequence name in 
				QNAME[x]=fgetc(fastq);		//QNAME and discard the rest 
			}					//of the information from the
			if (QNAME[x]!='\n')			//first line if any additional
			{					//information exists.
				fgets(exclude, 200, fastq);	
				QNAME[x]='\n';
			}
			QNAME[x+1]='\0';
			x=0;
			seq[x]=fgetc(fastq);			//Line 2: sequence of short read
			while (seq[x] != '\n')			//Read in sequence and convert 
			{					//bases to numerical values
				x++;
				seq[x]=fgetc(fastq);
			}
			read_length=x;
			fgets(exclude,200, fastq);		//Ignore Line 3:
			fgets(qualities,MRL, fastq);		//Line 4: Read qualities.
			temp_mismatches_allowed=*(mismatch_holder[reads_processed/4000000]+(reads_processed%4000000));
			reads_processed++;
			total_hits[0]=0;
			for (p=3; p>=0; p--)
			  {
				if (strcmp(ACGT_names[p][0],QNAME)==0)
				{
					if (ACGT_mismatches[p][0] <= temp_mismatches_allowed)
					{
						temp_mismatches_allowed=ACGT_mismatches[p][0];
						for(c=0; c<ACGT_hits[p][0];c++)
						{
							already_found=false;
							for (int count=0; count<total_hits[0] && already_found==false; count++)
							{
								if (all_locs[0][0][count]==ACGT_locs[p][0][c])
									already_found=true;
							}
							if (already_found==false)
							{
							  all_locs[0][0][total_hits[0]]=ACGT_locs[p][0][c];
							  all_locs[1][0][total_hits[0]]=ACGT_strand[p][0][c];
							  total_hits[0]++;
							}
						}
					}	
					else p=-1;
				}
			}	
	
			if (total_hits[0]!=0)
			{
				prints=0;
				if (total_hits[0]<max_reads_displayed)
					to_print=total_hits[0];
				else 	to_print=max_reads_displayed;
				while (prints<to_print)
				{
					rand_read=rand()%total_hits[0];
					for(x=0; QNAME[x] != '\n'; x++)
						fprintf(alignment, "%c",QNAME[x]);
					fprintf( alignment, "\t");
					j=0;
					while(chr_ref_len[j] < all_locs[0][0][rand_read])
					{
						all_locs[0][0][rand_read]-=chr_ref_len[j];
						j++;
					}
					fprintf(alignment, "%u\t%s\t%u\t",all_locs[1][0][rand_read], chr_name[j], all_locs[0][0][rand_read]);
					if (total_hits[0]==1 && prints==0)
						fprintf(alignment, "%d0\t%uM\t*\t0\t0\t",pHRED_table[read_length][temp_mismatches_allowed], read_length);
					else 	fprintf(alignment, "0\t%uM\t*\t0\t0\t", read_length);
					for (x=0; x<read_length; x++)
						fprintf(alignment, "%c",seq[x]);
					fprintf(alignment, "\t");
					for (x=0; x<read_length; x++)
						fprintf(alignment, "%c", qualities[x]);	
					fprintf(alignment, "\n");
					all_locs[0][0][rand_read]=all_locs[0][0][total_hits[0]-1];
					all_locs[1][0][rand_read]=all_locs[1][0][total_hits[0]-1];
					total_hits[0]--;
					prints++;
				}
			}
			else if (dump_unmatched==true)
				fprintf(fq_out[0], "@%s%s+%s%s",QNAME,seq,QNAME,qualities);
			for (p=3; p>=0; p--)
			{
				if (strcmp(ACGT_names[p][0],QNAME)==0)
				{
					fgets(ACGT_names[p][0],200,ACGT_alignment[p][0]);
					fread(&ACGT_hits[p][0],4,1,ACGT_alignment[p][0]);
					fread(&ACGT_mismatches[p][0],4,1,ACGT_alignment[p][0]);
					fread(&ACGT_locs[p][0],4,ACGT_hits[p][0],ACGT_alignment[p][0]);
					fread(&ACGT_strand[p][0],4,ACGT_hits[p][0],ACGT_alignment[p][0]);
				}
			}
			if ((reads_processed%1000000)==0)
			  fprintf(stdout, "%d reads processed for permanent file.\n",reads_processed);
		}
		fclose(fastq);
		}
	}
	else
	{
	  reads_processed=0;
		for (n=3; n<stop; n+=2)
		{
		fastq=fopen64(argv[n], "r");
		fastq2=fopen64(argv[n+1], "r");
		while (is_amp=fgetc(fastq) != EOF)
		{
		  reads_processed++;
			is_amp=fgetc(fastq2);
			x=0;
			QNAME[x]=fgetc(fastq);
			while (QNAME[x] != ' ' && QNAME[x] != '\t' && QNAME[x] != '\n')
			{					//First line of a short read:
				x++;				//Store the sequence name in 
				QNAME[x]=fgetc(fastq);		//QNAME and discard the rest 
			}					//of the information from the
			if (QNAME[x]!='\n')			//first line if any additional
			{					//information exists.
				fgets(exclude, 200, fastq);	
				QNAME[x]='\n';
			}
			QNAME[x+1]='\0';
			x=0;
			seq[x]=fgetc(fastq);			//Line 2: sequence of short read
			while (seq[x] != '\n')			//Read in sequence and convert 
			{					//bases to numerical values
				x++;
				seq[x]=fgetc(fastq);
			}
			read_length=x;
			fgets(exclude,200, fastq);		//Ignore Line 3:
			fgets(qualities,MRL, fastq);		//Line 4: Read qualities.
			//Read 2 in the pair.
			x=0;
			QNAME2[x]=fgetc(fastq2);
			while (QNAME2[x] != ' ' && QNAME2[x] != '\t' && QNAME2[x] != '\n')
			{					//First line of a short read:
				x++;				//Store the sequence name in 
				QNAME2[x]=fgetc(fastq2);		//QNAME and discard the rest 
			}					//of the information from the
			if (QNAME2[x]!='\n')			//first line if any additional
			{					//information exists.
				fgets(exclude, 200, fastq2);	
				QNAME2[x]='\n';
			}
			QNAME2[x+1]='\0';
			x=0;
			seq2[x]=fgetc(fastq2);			//Line 2: sequence of short read
			while (seq2[x] != '\n')			//Read in sequence and convert 
			{					//bases to numerical values
				x++;
				seq2[x]=fgetc(fastq2);
			}
			rl2=x;
			fgets(exclude,200, fastq2);		//Ignore Line 3:
			fgets(qualities2,MRL, fastq2);		//Line 4: Read qualities.
			
			//Pair 1
			total_hits[0]=0;
			for (p=3; p>=0; p--)
			{
				if (strcmp(ACGT_names[p][0],QNAME)==0)
				{
					for(c=0; c<ACGT_hits[p][0];c++)
					{
						already_found=false;
						for (int count=0; count<total_hits[0] && already_found==false; count++)
						{
							if (all_locs[0][0][count]==ACGT_locs[p][0][c])
								already_found=true;
						}
						if (already_found==false)
						{
							all_locs[0][0][total_hits[0]]=ACGT_locs[p][0][c];
							all_locs[1][0][total_hits[0]]=ACGT_strand[p][0][c];
							all_locs[2][0][total_hits[0]]=ACGT_mismatches[p][0];
							total_hits[0]++;
						}
					}
				}	
			}	
			//Pair 2
			total_hits[1]=0;
			for (p=3; p>=0; p--)
			{
				if (strcmp(ACGT_names[p][1],QNAME2)==0)
				{
					for(c=0; c<ACGT_hits[p][1];c++)
					{
						already_found=false;
						for (int count=0; count<total_hits[1] && already_found==false; count++)
						{
							if (all_locs[0][1][count]==ACGT_locs[p][1][c])
								already_found=true;
						}
						if (already_found==false)
						{
							all_locs[0][1][total_hits[1]]=ACGT_locs[p][1][c];
							all_locs[1][1][total_hits[1]]=ACGT_strand[p][1][c];
							all_locs[2][1][total_hits[1]]=ACGT_mismatches[p][1];
							total_hits[1]++;
						}
					}
				}	
			}
			match_num=0;
			if (total_hits[0] !=0 && total_hits[1] !=0)
			{
				for (b=0; b<total_hits[0]; b++)
				{
					for (c=0; c<total_hits[1]; c++)
					{
						if (all_locs[0][0][b] >= all_locs[0][1][c])
						{
							if (all_locs[0][0][b] - all_locs[0][1][c] <=max_insert_size && match_num<4*ALLOWED)
							{
								matched_pairs[match_num][0]=b;
								matched_pairs[match_num][1]=c;
								match_num++;
							}
						}
						else 	
							if (all_locs[0][1][c] - all_locs[0][0][b] <=max_insert_size && match_num<4*ALLOWED)
							{
								matched_pairs[match_num][0]=b;
								matched_pairs[match_num][1]=c;
								match_num++;
							}
					}
				}
			}
			if (match_num !=0)
			{
				p=read_length+rl2;
				for (b=0; b< match_num; b++)
				{
					if ((x=all_locs[2][0][matched_pairs[match_num][0]]+all_locs[2][1][matched_pairs[match_num][1]]) <= p)
					{
						if (x==p)
							prints=1;
						else 	
						{
							prints=0;
							p=x;
						}
						c=b;
					}
				}
				//Print mate 1.
				for(x=0; QNAME[x] != '\n'; x++)
					fprintf(alignment, "%c",QNAME[x]);
				fprintf( alignment, "\t");
				j=0;
				while(chr_ref_len[j] < all_locs[0][0][matched_pairs[c][0]])
				{
					all_locs[0][0][matched_pairs[c][0]]-=chr_ref_len[j];
					j++;
				}
				fprintf(alignment, "%u\t%s\t%u\t",all_locs[1][0][matched_pairs[c][0]]+64, chr_name[j], all_locs[0][0][matched_pairs[c][0]]);
				if (prints==0)
					fprintf(alignment, "%d0\t%uM\t*\t0\t0\t",pHRED_table[read_length+rl2][p], read_length);
				else 	fprintf(alignment, "0\t%uM\t*\t0\t0\t", read_length);
				for (x=0; x<read_length; x++)
					fprintf(alignment, "%c",seq[x]);
				fprintf(alignment, "\t");
				for (x=0; x<read_length; x++)
					fprintf(alignment, "%c", qualities[x]);	
				fprintf(alignment, "\n");
				
				//Print mate 2.
				for(x=0; QNAME2[x] != '\n'; x++)
					fprintf(alignment, "%c",QNAME2[x]);
				fprintf( alignment, "\t");
				j=0;
				while(chr_ref_len[j] < all_locs[0][1][matched_pairs[c][1]])
				{
					all_locs[0][1][matched_pairs[c][1]]-=chr_ref_len[j];
					j++;
				}
				fprintf(alignment, "%u\t%s\t%u\t",all_locs[1][1][matched_pairs[c][1]]+128, chr_name[j], all_locs[0][1][matched_pairs[c][1]]);
				if (prints==0)
					fprintf(alignment, "%d0\t%uM\t*\t0\t0\t",pHRED_table[read_length+rl2][p], rl2);
				else 	fprintf(alignment, "0\t%uM\t*\t0\t0\t", rl2);
				for (x=0; x<rl2; x++)
					fprintf(alignment, "%c",seq2[x]);
				fprintf(alignment, "\t");
				for (x=0; x<rl2; x++)
					fprintf(alignment, "%c", qualities2[x]);	
				fprintf(alignment, "\n");
			}
			else if (dump_unmatched==true)
			{
				fprintf(fq_out[0], "@%s%s+%s%s",QNAME,seq,QNAME,qualities);
				fprintf(fq_out[1], "@%s%s+%s%s",QNAME2,seq2,QNAME2,qualities2);
			}
			for (p=3; p>=0; p--)
			{
				if (strcmp(ACGT_names[p][0],QNAME)==0)
				{
					fgets(ACGT_names[p][0],200,ACGT_alignment[p][0]);
					fread(&ACGT_hits[p][0],4,1,ACGT_alignment[p][0]);
					fread(&ACGT_mismatches[p][0],4,1,ACGT_alignment[p][0]);
					fread(&ACGT_locs[p][0],4,ACGT_hits[p][0],ACGT_alignment[p][0]);
					fread(&ACGT_strand[p][0],4,ACGT_hits[p][0],ACGT_alignment[p][0]);
				}
				if (strcmp(ACGT_names[p][1],QNAME2)==0)
				{
					fgets(ACGT_names[p][1],200,ACGT_alignment[p][1]);
					fread(&ACGT_hits[p][1],4,1,ACGT_alignment[p][1]);
					fread(&ACGT_mismatches[p][1],4,1,ACGT_alignment[p][1]);
					fread(&ACGT_locs[p][1],4,ACGT_hits[p][1],ACGT_alignment[p][1]);
					fread(&ACGT_strand[p][1],4,ACGT_hits[p][1],ACGT_alignment[p][1]);
				}
			}
			if ((reads_processed%1000000)==0)
			  fprintf(stdout, "%d reads processed for permanent file\n",reads_processed);
		}
		fclose(fastq);
		}

	}
	for (p=0; p<4;p++)
	{
		fclose(ACGT_alignment[p][0]);
		strcpy(fullname,basename);
		if 	(p==0)
			strcat(fullname, ".A");
		else if	(p==1)
			strcat(fullname, ".C");
		else if	(p==2)
			strcat(fullname, ".G");
		else	strcat(fullname, ".T");
		if (remove(fullname) !=0)
			fprintf(stdout, "error removing file %s.\n",fullname);
		if (PAIRS==2)
		{
			strcat(fullname, ".2");
			if (remove(fullname) != 0)
				fprintf(stdout, "error removing file %s.\n", fullname);
		}
	}
	fprintf(stdout, "%d reads processsed for permanent file\n",reads_processed);
		
	fclose(index);
	fclose(alignment);
 	return 0;
 }
int usage()
{
	fprintf(stderr, "\nUsage: \talign <index.sqn> { file1.fastq ... fileN.fastq } <alignment.sam> [options]\n\n");
	fprintf(stderr, "Options:\t-p int\tPrint a maximum of p reads[1].\n");
	fprintf(stderr, "        \t-m int\tLimit alignment to at most m mismatches.\n");
	fprintf(stderr, "        \t-a int\tMaximum number of alignments with equal number of mismatches to consider per quater [5].\n");
	fprintf(stderr, "        \t-r int\tMaxiumum length of reads in fastq file [1000].\n");
	fprintf(stderr, "        \t-q int\tOnly search for alignments with a quality of q or higher.\n");
	fprintf(stderr, "        \t-g ull\tDefine the length of reference genome.\n");
	fprintf(stderr, "        \t-s int\tOnly consider a maximum of s index locations per hit. -1 for all locations [100].\n");
	fprintf(stderr, "        \t-P    \tPair end alignment mode. Imput fastqs in format file1.1.fq file 1.2.fq ...fileN.1.fq fileN.2.fq\n");
	fprintf(stderr, "        \t-i int\tMaximum insert size allowed between two mates in a pair [1000]\n");
	fprintf(stderr, "        \t-f str\tWrite unmaligned reads to file. [Disabled by default].\n");
	return 1;
}
